import json
from pathlib import Path
import os
import signal
from typing import TYPE_CHECKING
import multiprocessing
import click
from pydantic import ValidationError
import uvicorn
from locknessie.settings import safely_get_settings, OpenIDIssuer, NoConfigError, get_config_path
import functools
from locknessie.settings import ConfigSettings

if TYPE_CHECKING:
    from pathlib import Path

def add_config_options(command):
    for field_name, field in ConfigSettings.model_fields.items():
        if field_name == "config_path":
            continue
        option_name = f"--{field_name.replace('_', '-')}"
        option_kwargs = {"default": None, "help": f"Config option for {field_name}"}
        if hasattr(field, 'annotation'):
            option_kwargs["type"] = field.annotation if field.annotation in [int, str, bool] else str
        command = click.option(option_name, **option_kwargs)(command)
    return command

def _filter_none_kwargs(kwargs):
    return {k: v for k, v in kwargs.items() if v is not None}

def _sync_config_settings(initial_config_dict: dict, config_file: "Path") -> None:
    """round-about way to populate the config defaults to make user input easier
    by exposing them in the config file.

    Args:
        initial_config_dict: the required values that the config object needs to be populated with
        config_file: the path to the config file to be written to
    """
    # hack to deal with serialization issues
    for key, value in initial_config_dict.items():
        if isinstance(value, Path):
            initial_config_dict[key] = str(value.absolute())

    config_file.write_text(json.dumps(initial_config_dict, indent=4))
    synced_config_settings = safely_get_settings().model_dump_json(indent=4)
    config_file.write_text(synced_config_settings)

def _start_server(port: int, host: str):
    """Start the HTTP service."""
    from locknessie.server import app
    uvicorn.run(app, host=host, port=port)

def _stop_server(pid: int, sig: int):
    """Stop the HTTP service."""
    os.kill(pid, sig)


@click.group()
def cli():
    """Lock Nessie CLI"""

@cli.group()
def config():
    """Configure Lock Nessie settings."""

@config.command()
def init():
    """Initialize a new config file"""
    config_dict = {}
    config_path = get_config_path()
    if config_path.exists():
        click.echo(f"Config file already exists at {config_path}")
        return
    config_path.parent.mkdir(parents=True, exist_ok=True)

    config_dict["config_path"] = config_path
    config_dict["config_dir"] = config_path.parent
    config_dict["openid_issuer"] = click.prompt("Which OpenID provider are you using?", type=click.Choice([v.value for v in OpenIDIssuer]))
    match config_dict["openid_issuer"]:
        case OpenIDIssuer.microsoft:
            config_dict["openid_client_id"] = click.prompt("What is your Microsoft client ID?", type=str)
            if click.prompt("Do you want to set daemon auth? (y/n)", type=click.Choice(["y", "n"])) == "y":
                config_dict["openid_secret"] = click.prompt("What is your Microsoft Application secret?", type=str)
                config_dict["openid_tenant"] = click.prompt("What is your Microsoft tenant ID?", type=str)
        case _:
            click.secho("That provider is not supported yet", fg="red", err=True)
            return
    _ = _sync_config_settings(config_dict, config_path)
    click.echo(f"Config file initialized at {config_path}.")

def catch_noconfigerror(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except NoConfigError as e:
            click.secho(f"Error: {e}", fg="red", err=True)
            return
    return wrapper

def catch_invalid_config(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ValidationError as e:
            error_list = e.errors()
            if error_list:
                click.secho("Invalid or missing config values:", fg="red", err=True)
                for err in error_list:
                    loc = ".".join(str(x) for x in err.get("loc", []))
                    msg = err.get("msg", "Unknown error")
                    click.secho(f"  - {loc}: {msg}", fg="yellow", err=True)
                click.echo("If using the CLI or browser auth, please run `locknessie config init` to initialize the config file.")
            else:
                click.secho(f"Error: {e}", fg="red", err=True)
    return wrapper

@config.command()
@click.argument("key")
@click.argument("value")
@catch_noconfigerror
def set(key: str, value: str):
    """Set a config value"""
    config_settings = safely_get_settings()
    try:
        setattr(config_settings, key, value)
    except AttributeError:
        click.echo(f"Invalid config key: {key}")
        return
    _ = _sync_config_settings(config_settings.model_dump(), config_settings.config_path)
    click.echo(f"Config file updated at {config_settings.config_path}.")


@cli.group()
@add_config_options
@click.pass_context
def service(ctx, **kwargs):
    """Manage the Lock Nessie HTTP service."""
    ctx.obj = _filter_none_kwargs(kwargs)
    raise NotImplementedError(("Service will allow polling to auth tools like Dremio. "
                              "Sevice implementation is not yet complete, stay tuned!"))

@service.command()
@click.option("--daemon", is_flag=True, help="Run the HTTP service as a daemon")
@click.pass_context
@catch_noconfigerror
def start(ctx, daemon):
    """Start the HTTP service."""
    config_kwargs = ctx.obj or {}
    config_settings = ConfigSettings(**config_kwargs)
    port = config_settings.auth_callback_port
    host = config_settings.auth_callback_host
    if daemon:
        process = multiprocessing.Process(target=_start_server, args=(port, host))
        pid = process.start()
        pid_cache = config_settings.config_path / "pid"
        pid_cache.write_text(str(pid))
        click.echo(f"HTTP service started on {host}:{port} with PID {pid}")
    else:
        _start_server(port, host)

@service.command()
@click.option("--kill", is_flag=True, help="Send a SIGKILL signal to the HTTP service. Useful if the server refuses to stop.")
@click.pass_context
@catch_noconfigerror
def stop(ctx, kill: bool):
    """Stop the HTTP service."""
    config_kwargs = ctx.obj or {}
    config_settings = ConfigSettings(**config_kwargs)
    sig_enum = "SIGKILL" if kill else "SIGINT"
    sig = getattr(signal, sig_enum)
    pid_cache = config_settings.config_path / "pid"
    if not pid_cache.exists():
        click.echo("HTTP service is not running")
        return
    pid = int(pid_cache.read_text())
    click.echo(f"Terminating HTTP service with PID {pid} with signal {sig_enum}")
    _stop_server(pid, sig)
    click.echo(f"HTTP service {sig_enum} stop signal sent")

@cli.group()
@add_config_options
@click.pass_context
def token(ctx, **kwargs):
    """Manage OpenID tokens."""
    ctx.obj = _filter_none_kwargs(kwargs)

@token.command()
@click.pass_context
@catch_noconfigerror
@catch_invalid_config
def show(ctx):
    """Show the active OpenID bearer token."""
    from locknessie.main import LockNessie
    config_kwargs = ctx.obj or {}
    settings = safely_get_settings(**config_kwargs)
    import locknessie.settings as settings_mod
    settings_mod.settings = settings
    token = LockNessie().get_token()
    click.echo(token)





from typing import TYPE_CHECKING
from locknessie.settings import safely_get_settings, OpenIDIssuer
from typing import Optional
from locknessie.auth_providers.base import AuthType

if TYPE_CHECKING:
    from locknessie.settings import ConfigSettings

class LockNessie:
    settings: "ConfigSettings"

    def __init__(self, auth_type: Optional[str] = AuthType.user, **kwargs):
        """set the correct provider based on the settings"""
        self.settings = safely_get_settings(**kwargs)
        self.provider = self._get_provider(auth_type=auth_type)

    def _get_provider(self, auth_type: str) -> str:
        """returns the correct provider based on the settings"""
        match self.settings.openid_issuer:
            case OpenIDIssuer.microsoft:
                from locknessie.auth_providers.microsoft import MicrosoftAuth
                return MicrosoftAuth(auth_type=auth_type)
            case _:
                raise NotImplementedError("Not implemented")

    def get_token(self) -> str:
        """returns the authed/updated bearer token to be used for the OpenID connection"""
        return self.provider.get_token()


from typing import TYPE_CHECKING, Optional, Union
import msal
from locknessie.logger import get_logger
from locknessie.settings import safely_get_settings
from locknessie.auth_providers.base import AuthBase, AuthType

if TYPE_CHECKING:
    from pathlib import Path

settings = safely_get_settings()

logger = get_logger(__name__)

class MicrosoftAuth(AuthBase):
    scopes: list[str]
    cache_file_name: str = "microsoft/cache.bin"
    cache_file: "Path"
    account: "msal.Account"
    app: "msal.PublicClientApplication"
    cache: "msal.SerializableTokenCache"

    def __init__(self,
                 auth_type: Optional[AuthType] = AuthType.user):
        """
        auth_type: the type of authentication to use - users can log in with a browser, daemons can use secret credentials.
        """
        super().__init__(auth_type)
        self.cache = None
        if self.auth_type == AuthType.user:
            self.scopes = ["User.Read"]
            self.cache_file = self.initilaize_cache_file(self.cache_file_name)
            self.cache = self._load_cache(self.cache_file)
        elif self.auth_type == AuthType.daemon:
            self.scopes = ["https://graph.microsoft.com/.default"]
            assert settings.openid_secret, "OpenID secret is required for daemon auth"
        else:
            raise ValueError(f"Invalid auth type: {self.auth_type}")

        self.app = self._get_app(self.cache)
        self.account = self._get_client_id_account(self.app)


    def get_token(self) -> str:
        return self._get_token()

    def initilaize_cache_file(self, cache_file_name: str) -> "Path":
        cache_file = settings.config_dir / cache_file_name
        cache_file.parent.mkdir(parents=True, exist_ok=True)
        return cache_file

    def _load_cache(self, cache_file: "Path") -> msal.SerializableTokenCache:
        logger.info("initializing auth cache...")
        cache = msal.SerializableTokenCache()
        if cache_file.exists():
            logger.info("loading auth cache from %s", cache_file)
            cache.deserialize(cache_file.read_text())
            logger.info("auth cache loaded")
        logger.info("auth cache initialized")
        return cache

    def _get_app(self, cache: Optional["msal.SerializableTokenCache"] = None) -> "msal.PublicClientApplication":
        match self.auth_type:
            case AuthType.daemon:
                return msal.ConfidentialClientApplication(
                    client_id=settings.openid_client_id,
                    client_credential=settings.openid_secret,
                    authority=f"https://login.microsoftonline.com/{settings.openid_tenant}"
                )
            case AuthType.user:
                extra_args = {}
                if not settings.openid_allow_all_tenants:
                    extra_args["authority"] = f"https://login.microsoftonline.com/{settings.openid_tenant}"
                return msal.PublicClientApplication(client_id=settings.openid_client_id,
                                                    token_cache=cache,
                                                    **extra_args)
            case _:
                raise ValueError(f"Invalid auth type: {self.auth_type}")

    def _get_client_id_account(self, app: "msal.PublicClientApplication") -> "msal.Account":
        """get the first selected account. Note that browser auth needs to succeed first."""
        try:
            return app.get_accounts()[0]
        except IndexError:
            return None

    def _get_token(self) -> str:
        match self.auth_type:
            case AuthType.user:
                return self._get_user_token()
            case AuthType.daemon:
                return self._get_daemon_token()
            case _:
                raise ValueError(f"Invalid auth type: {self.auth_type}")

    def _get_daemon_token(self) -> str:
        logger.info("getting daemon token...")
        result = self.app.acquire_token_for_client(scopes=self.scopes)
        return self._extract_token_from_result(result)

    def _get_user_token(self) -> str:
        logger.info("getting user token...")
        logger.info("attempting to get token silently with refresh...")
        if not self.account or \
            not (result := self.app.acquire_token_silent(scopes=self.scopes, account=self.account)):
            logger.info("no token or account found, attempting to get token interactively via browser...")
            result = self.app.acquire_token_interactive(scopes=self.scopes, port=1234)
        token = self._extract_token_from_result(result)
        self._save_cache(self.cache_file, self.cache)
        return token

    def _extract_token_from_result(self, result: dict) -> str:
        if "access_token" in result:
            logger.info("token acquired")
            return result["access_token"]
        msg = result.get("error_description", "Authentication failed with an unknown error")
        raise ValueError(msg)

    def _save_cache(self, cache_file: "Path", cache: "msal.SerializableTokenCache"):
        cache_file.write_text(cache.serialize())





from enum import Enum
from pathlib import Path
import os
from typing import Union, Optional, Type
from pydantic import Field, ValidationError, model_validator
from pydantic_settings import BaseSettings, PydanticBaseSettingsSource, JsonConfigSettingsSource, InitSettingsSource, SettingsConfigDict
from typing import Optional

from locknessie.logger import get_logger

logger = get_logger(__name__)

class NoConfigError(Exception):
    """exception raised when no config file is found"""
    pass

class Environment(str, Enum):
    release = "release" # code that is read-only and published in a release
    development = "development" # code that is being actively developed
    production = "release" # legacy value for backwards compatibility

    def __eq__(self, other):
        """legacy support for the value `production` which is confusing to users"""
        if isinstance(other, str):
            if self is Environment.release and other == "production":
                return True
            return str(self.value) == other
        if isinstance(other, Environment):
            return str(self.value) == str(other.value)
        return NotImplemented

class OpenIDIssuer(str, Enum):
    microsoft = "microsoft"
    keycloak = "keycloak"

def get_config_path(config_path: Optional[Union[Path, str]] = None) -> Path:
    """gets the config _json file_ full path"""
    if config_path:
        logger.debug(f"Using config path from CLI: {config_path}")
        config_path = Path(config_path)

    elif from_env := os.getenv("LOCKNESSIE_CONFIG_PATH"):
        logger.debug(f"Using config path from env: {from_env}")
        config_path = Path(from_env)
    else:
        logger.debug("Using default config path: ~/.locknessie/config.json")
        config_path = Path.home() / ".locknessie" / "config.json"
    assert not config_path.is_dir(), "Config path must be a file, not a directory"
    assert config_path.suffix == ".json", "Config path must be a JSON file"
    return config_path



class ConfigSettings(BaseSettings):
    """settings derived from the cli, envars, and config file"""

    model_config = SettingsConfigDict(
        json_file_encoding="utf-8",
        env_prefix="locknessie_",
        case_sensitive=False
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: InitSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        # Precedence: CLI/init > env > config file
        return (
            init_settings,  # CLI args
            env_settings,   # Environment variables
            JsonConfigSettingsSource(settings_cls, init_settings.init_kwargs["config_path"]),  # recursively load config file
        )

    config_path: Path
    config_dir: Path = Field(..., description="The directory for the config file")
    environment: Environment = Field(Environment.release, description="'release' for released code, 'development' for local development")
    openid_issuer: OpenIDIssuer = Field(..., description="The issuer of the OpenID client")
    openid_client_id: str = Field(..., description="The client ID of the OpenID client")
    openid_tenant: Optional[str] = Field(None, description="The tenant of the OpenID client")
    openid_secret: Optional[str] = Field(None, description="The secret of the OpenID client, used for daemon auth")
    openid_realm: Optional[str] = Field(None, description="The realm of the OpenID client")
    openid_url: Optional[str] = Field(None, description="The URL of the OpenID provider")

    # Secret settings
    secret_name: str = Field("token", description="The name of the secret to be retrieved")

    # port settings
    auth_callback_port: Optional[int] = Field(default=1234, description="The port for the OpenID auth callback server")
    impersonation_port: Optional[int] = Field(default=8200, description="The port for the vault impersonation server")
    auth_callback_host: Optional[str] = Field(default="0.0.0.0", description="The host for the OpenID auth callback server")

    @model_validator(mode="before")
    def set_config_dir_from_path(cls, values):
        if values.get("config_path") and not values.get("config_dir"):
            values["config_dir"] = Path(values["config_path"]).parent
        return values

    @model_validator(mode="before")
    def normalize_environment(cls, values):
        env = values.get("environment")
        if env == "production":
            values["environment"] = "release"
        return values

def safely_get_settings(config_path: Optional[Union[Path, str]] = None,
                        **kwargs) -> Type["BaseSettings"]:
    """safely get the settings and direct the user to init the config if needed
    """
    if config_path:
        try:
            assert config_path.exists()
        except AssertionError as e:
            raise NoConfigError(
                f"Config file not found at {config_path}. Please run `locknessie config init` to initialize the config file."
            ) from e

    config_path = get_config_path(config_path)
    return ConfigSettings(config_path=config_path, **kwargs)




import datetime
import pytz
from fastapi import FastAPI
from fastapi.responses import JSONResponse
from fastapi.exceptions import HTTPException
from locknessie.settings import safely_get_settings
import locknessie.main as main

settings = safely_get_settings()

app = FastAPI()

@app.get("/v1/secret/data/{secret_name}")
async def get_secret(secret_name: str):
    try:
        assert secret_name == settings.secret_name
    except AssertionError:
        raise HTTPException(status_code=400, detail="Invalid secret name")
    token = main.LockNessie().get_token()
    now = datetime.datetime.now(pytz.utc).isoformat() + "Z"
    response = {
        "data": {
            "data": {
                secret_name: token
            },
            "metadata": {
                "created_time": now,
                "custom_metadata": None,
                "deletion_time": "",
                "destroyed": False,
                "version": 1
            }
        }
    }
    return JSONResponse(content=response)




from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional

class AuthType(str, Enum):
    user = "user"
    daemon = "daemon"


class AuthBase(ABC):
    _auth_type: AuthType

    def __init__(self, auth_type: Optional[AuthType] = AuthType.user):
        self._auth_type = auth_type

    @property
    def auth_type(self) -> AuthType:
        return self._auth_type

    @abstractmethod
    def get_token(self) -> str:
        pass